#include <stdio.h>
#include <stdlib.h>
#include "types.h"
#include "polygon.h"
#include "vl.h"

extern S32 FillConvexPolygon(struct PointListHeader *VertexList ,S32 Colour, S32 XOffset, S32 YOffset);

/* Matrix multiplies Xform by SourceVec and stores the result in */
/* DestVec. Multiplies a 4 x 4 matrix times a 4 x 1 matrix; the  */
/* result is a 4 x 1 matrix, as follows:                         */
/* --   --    --  --   --  --                                    */
/* |     |     | 4 |    | 4 |                                    */ 
/* | 4x4 |  x  | x | =  | x |                                    */
/* |     |     | 1 |    | 1 |                                    */
/* --   --    --  --   --  --                                    */
void XformVec(double Xform[4][4], double *SourceVec, double *DestVec)
{
    S32 i;
    S32 j;
    
    for(i = 0; i < 4; i++)
    {
        DestVec[i] = 0;  
        for(j = 0; j < 4; j++)
        {
            DestVec[i] += Xform[i][j] * SourceVec[j];
        }     
    }     
}

/* Matrix multiplies SourceXform1 by SourceXform2 and stores the */
/* result in DestXform. Multiplies a 4 x 4 matrix times a 4 x 4  */
/* matrix: the result is a 4 x 4 matrix, as follows:             */
/* --   --     --     --   --  --                                */
/* |     |     |       |    | 4 |                                */ 
/* | 4x4 |  x  | 4 x 4 | =  | x |                                */
/* |     |     |       |    | 1 |                                */
/* --   --     --     --   --  --                                */
void ConcatXforms(const double SourceXform1[4][4], const double SourceXform2[4][4],double (*DestXform)[4])
{
    S32 i;
    S32 j;
    S32 k;
    
    for(i = 0; i < 4; i++)
    {
        for(j=0 ; j < 4; j++)
        {
            /* clear dest before starting matrix concat*/
            DestXform[i][j] = 0;
            /* concat matrixs */
            for(k = 0; k < 4; k++)
            {
                DestXform[i][j] += SourceXform1[i][k] * SourceXform2[k][j];
            }
        }
    } 
}

/* Transforms convex polygon poly (which has polylength vertices)    */
/* performing the transformation according to Xform (which generally */
/* represents a transformation from object space through world space */
/* to view space), then projects the transformed polygon onto the    */
/* screen and draws it in colour Colour.                             */
void XformAndProjectPoly(double Xform[4][4], struct Point3 *Poly, S32 PolyLength, S32 Colour)
{
    S32 i;
    struct Point3 XformedPoly[MAX_POLY_LENGTH];
    struct Point  ProjectedPoly[MAX_POLY_LENGTH];
    struct PointListHeader Polygon;
    
    /* transforms to view space, then project to the screen */
    for(i = 0; i < PolyLength; i++)
    {
        /* transform to view space */
        XformVec(Xform, (double *)&Poly[i], (double *)&XformedPoly[i]);
        /* project the X & Y coordinates to the screen, rounding to the  */
        /* nearest integral coordinates. The Y coordinates is negated to */
        /* flip from view space, where increasing Y is up, to screen     */
        /* space where increasing Y is down. Add in half the screen      */
        /* width and height to center on the screen                      */
        ProjectedPoly[i].X = ((S32)(XformedPoly[i].X / XformedPoly[i].Z * PROJECTION_RATIO * (SCREEN_WIDTH / 2.0) + 0.5)) + SCREEN_WIDTH/2; 
        ProjectedPoly[i].Y = ((S32)(XformedPoly[i].Y / XformedPoly[i].Z * -1.0 * PROJECTION_RATIO * (SCREEN_WIDTH / 2.0) + 0.5)) + SCREEN_HEIGHT/2;                
    }
       
    /* Draw the polygon */
    DRAW_POLYGON(ProjectedPoly, PolyLength, 4, 0, 0);  
}
